from datetime import datetime, timedelta
import os
import sys


ENOENT, EMEDIUMTYPE = range(-1, -3, -1)
PY3 = sys.version_info[0] == 3

##############################################################################
# Copied from:
# http://eli.thegreenplace.net/2011/10/19/perls-guess-if-file-is-text-or-binary-implemented-in-python
##############################################################################

# A function that takes an integer in the 8-bit range and returns
# a single-character byte object in py3 / a single-character string
# in py2.
#
int2byte = (lambda x: bytes((x,))) if PY3 else chr


def istextfile(block):
    """ Uses heuristics to guess whether the given file is text or binary,
        by reading a single block of bytes from the file.
        If more than 30% of the chars in the block are non-text, or there
        are NUL ('\x00') bytes in the block, assume this is a binary file.
    """
    if b'\x00' in block:
        # Files with null bytes are binary
        return False
    elif not block:
        # An empty file is considered a valid text file
        return True

    _text_characters = (
        b''.join(int2byte(i) for i in range(32, 127)) +
        b'\n\r\t\f\b'
    )
    # Use translate's 'deletechars' argument to efficiently remove all
    # occurrences of _text_characters from the block
    nontext = block.translate(None, _text_characters)
    return float(len(nontext)) / len(block) <= 0.30


def parse_date(startdate_str):
    """parse the date SAR collection started"""
    formats = ['%y-%m-%d', '%Y-%m-%d', '%m/%d/%y']
    for fmt in formats:
        try:
            startdate = datetime.strptime(startdate_str, fmt)
            return startdate
        except ValueError:
            continue
    else:  # pylint: disable=W0120
        raise Exception('Unknown date format: %s' % startdate_str)


def format_time(startdate, timestr, afternoon):
    """format the timstr as a datetime str"""
    datetime_str = '%s-%s-%s %s %s' % (startdate.year, startdate.month,
                                       startdate.day, timestr, afternoon)
    timedate = datetime.strptime(datetime_str, '%Y-%m-%d %I:%M:%S %p')
    return datetime.strftime(timedate, '%Y-%m-%dT%H:%M:%S')


def parsefile(filename, section, stats, key=''):
    """parses <filename>, pulling out SAR data corresponding to <section>
    and <stats>"""
    retvals = []
    headermap = None

    if not os.path.exists(filename):
        return ENOENT

    def getstatsline():
        retstr = format_time(startdate, vals[0], vals[1])
        for stat in stats:
            retstr += ',%s' % vals[headermap[stat]]

        return retstr

    with open(filename, 'r') as _file:
        headerstr = ','.join(['#Time'] + stats)
        retvals.append(headerstr)

        header = _file.readline()
        if not istextfile(header[:512]):
            return EMEDIUMTYPE

        header = header.split()
        startdate = parse_date(header[3])
        starttime = None

        key1, key2 = section.split('|')
        pm = False
        for line in _file:
            if line == '\n' or 'Average' in line or 'Summary' in line:
                continue

            if not starttime and key1 not in line:
                continue

            vals = line.split()

            # keep track of date roll over
            if vals[1] == 'PM':
                if not pm:
                    pm = True
            elif vals[1] == 'AM' and pm:
                startdate = startdate + timedelta(days=1)
                pm = False

            if not starttime and (key1 != vals[2] and key2 != vals[3]):
                continue

            if not starttime:
                starttime = vals[0]
                headermap = {n: idx for (idx, n) in enumerate(vals)}
                continue

            if vals[0] == starttime and len(retvals) > 2:
                return retvals

            if starttime:
                if not key:
                    retvals.append(getstatsline())
                elif vals[2] == key:
                    retvals.append(getstatsline())

    return retvals



import argparse
import errno
from . import cparsar


DEFAULT_CPU_STATS = ['%idle']
DEFAULT_MEM_STATS = ['%memused']
DEFAULT_DISK_STATS = ['tps']
DEFAULT_NETDEV_STATS = ['rxpck/s', 'txpck/s']
DEFAULT_QUEUE_STATS = ['runq-sz']
DEFAULT_SWAP_STATS = ['%swpused']


# Section identifiers
CPU = 'CPU|%usr'
MEM = 'kbmemfree|kbmemused'
DISK = 'DEV|tps'
NETDEV = 'IFACE|rxpck/s'
QUEUE = 'runq-sz|plist-sz'
SWAP = 'kbswpfree|kbswpused'


class FileNotFound(Exception):
    """Exception raised when sar file isn't found"""
    pass


class IncompatibleFileType(Exception):
    """Exception raised when parsar can't parse the file (most likely means
    the user passed in the binary sa file)"""
    pass


class ParsarException(Exception):
    """General catch all exception"""
    pass


def _try_parse(filename, section, stats, **kwargs):
    """Attempt to parse the <filename>, raising appropriate
    exception if it fails."""
    try:
        resp = cparsar.parsefile(filename, section, stats, **kwargs)

        if isinstance(resp, list):
            return resp

        if resp == cparsar.ENOENT:
            raise FileNotFound('file "%s" does not exist' % filename)
        elif resp == cparsar.EMEDIUMTYPE:
            msg = 'incompatible file detected, parsar requires SAR text ' \
                  'files, perhaps "%s" is the binary version?'
            raise IncompatibleFileType(msg % filename)
        else:
            raise ParsarException('Unexpected err code returned: "%s"' % resp)

    except Exception as e:  # pylint: disable=W0703
        raise ParsarException(e)


class Parsar(object):
    """Parsar class"""

    def __init__(self, filename):
        self.filename = filename

    # pylint: disable=E1101
    def cpu(self, stats=DEFAULT_CPU_STATS):
        return _try_parse(self.filename, CPU, stats, key='all')

    def mem(self, stats=DEFAULT_MEM_STATS):
        return _try_parse(self.filename, MEM, stats)

    def disk(self, devname, stats=DEFAULT_DISK_STATS):
        return _try_parse(self.filename, DISK, stats, key=devname)

    def netdev(self, devname, stats=DEFAULT_NETDEV_STATS):
        return _try_parse(self.filename, NETDEV, stats, key=devname)

    def queue(self, stats=DEFAULT_QUEUE_STATS):
        return _try_parse(self.filename, QUEUE, stats)

    def swap(self, stats=DEFAULT_SWAP_STATS):
        return _try_parse(self.filename, SWAP, stats)


def get_args():
    """get command line arguments"""
    parser = argparse.ArgumentParser(description='Parse SAR data')
    subparsers = parser.add_subparsers(help='SAR Section of interest')

    subparser = subparsers.add_parser('cpu', help='CPU stats')
    subparser.add_argument('--cpustats', nargs='*')

    subparser = subparsers.add_parser('mem', help='Memory stats')
    subparser.add_argument('--memstats', nargs='*')

    subparser = subparsers.add_parser('disk', help='Disk stats')
    subparser.add_argument('--diskdev', required=True,
                           help='desired disk device')
    subparser.add_argument('--diskstats', nargs='+',
                           help='default: %s' % DEFAULT_DISK_STATS[0])

    subparser = subparsers.add_parser('netdev', help='Network device stats')
    subparser.add_argument('--iface', required=True,
                           help='desired network interface')
    subparser.add_argument('--netstats', nargs='*',
                           help='default: %s' % DEFAULT_NETDEV_STATS[0])

    subparser = subparsers.add_parser('queue',
                                      help='Queue length and load stats')
    subparser.add_argument('--queuestats', nargs='*')

    subparser = subparsers.add_parser('swap', help='Swap space stats')
    subparser.add_argument('--swapstats', nargs='*')

    parser.add_argument('filename', help='the SAR file to parse')
    return parser.parse_args()


def main():
    """main entry point"""
    args = get_args()
    p = Parsar(args.filename)
    if hasattr(args, 'cpustats'):
        stats = args.cpustats or DEFAULT_CPU_STATS
        result = p.cpu(stats=stats)
    elif hasattr(args, 'memstats'):
        stats = args.memstats or DEFAULT_MEM_STATS
        result = p.mem(stats=stats)
    elif hasattr(args, 'diskstats'):
        stats = args.diskstats or DEFAULT_DISK_STATS
        result = p.disk(args.diskdev, stats=stats)
    elif hasattr(args, 'netstats'):
        stats = args.netstats or DEFAULT_NETDEV_STATS
        result = p.netdev(args.iface, stats=stats)
    elif hasattr(args, 'queuestats'):
        stats = args.queuestats or DEFAULT_QUEUE_STATS
        result = p.queue(stats=stats)
    elif hasattr(args, 'swapstats'):
        stats = args.swapstats or DEFAULT_SWAP_STATS
        result = p.swap(stats=stats)

    try:
        for r in result:
            print(r)
    except IOError as e:
        if e.errno == errno.EPIPE:
            pass
        else:
            raise e



from distutils.core import setup

setup(
    name='parsar',
    packages=['parsar'],
    version='0.2.0',
    description='Python SAR data parser',
    author='Abe Friesen',
    author_email='abefriesen.af@gmail.com',
    url='https://github.com/doyshinda/parsar',
    license='MIT',

    keywords=['sar'],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',

        'Operating System :: POSIX :: Linux',

        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 3',
    ],
    entry_points={
        'console_scripts': [
            'parsar = parsar.parsar:main',
        ],
    }
)
